/***************************************************************************
 *   Copyright (C) 2009-2010 Virginia Tech Real-Time Systems Lab
 *   Written by Matthew Dellinger
 *   mdelling@vt.edu
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define VERSION "1.1.4"

#include <QApplication>
#include <QFileDialog>
#include <QObject>
#include <iostream>
#include <unistd.h>
#include <sys/mman.h>
#include "ui_Sched_Test_App.h"
#include "rt_test.h"

using namespace std;

void set_version(Ui_MainWindow *ui) {
	ui->versionLabel->setText(VERSION);
}

void print_usage() {
	cout	<< endl << "Console testing application for ChronOS"			<< endl
		<< "------------------------------------------------------ " 		<< endl
		<< "Usage:" 								<< endl 
		<< "  -s scheduler  The name of scheduler - RMA, EDF, DASA etc" 	<< endl
		<< "  -p            Enable priority inheritance"			<< endl
		<< "  -h            Enable HUA abort handlers"				<< endl
		<< "  -d            Enable deadlock prevention"				<< endl
		<< "  -f filename   The name of the file that has the task list"	<< endl
		<< "  -b            Enable the batch mode"				<< endl
		<< "  -c cpu-usage  The cpu-usage"					<< endl
		<< "  -e end-usage  The end cpu-usage (used in batch mode)"		<< endl
		<< "  -i interval   The interval for each iteration in batch"		<< endl
		<< "  -r run-time   The total time the test needs to run"		<< endl
		<< "                This should be the hyper-period (use -z to find)"	<< endl
		<< "  -l cs-length  Enable locking, and provide critical"		<< endl
		<< "                section length"					<< endl
		<< "  -n            Enable nested locking (as opposed to sequential)"	<< endl
		<< "  -v            Enable verbose mode"				<< endl
		<< "  -o            Enable output to a log file"			<< endl
		<< "  -x            Enable log in Excel format"				<< endl
		<< "  -g            Enable log in Gnuplot format"			<< endl
		<< "  -z            Don't run the test, just print the hyper-period"	<< endl
		<< endl
		<< "Examples:"							 	<< endl
		<< "    $sched_test_app -s RMA -f taskset/5t_nl -c 100 -r 20"		<< endl
		<< endl
		<< " OR $sched_test_app -s EDF -f tasknet/10t_nl -c 50 -l 10 -r 20 --a -p -h -d"
		<< endl									<< endl
		<< " NOTE: -a -p -h -d can also be written as -aphd"			<< endl;
}

static void qtMessageHandler(QtMsgType type, const char *msg) {
	switch(type) {
	case QtDebugMsg:
		fprintf(stderr, "Debug: %s\n", msg);
	case QtWarningMsg:
		fprintf(stderr, "Warning: %s\n", msg);
	case QtCriticalMsg:
		fprintf(stderr, "Critical: %s\n", msg);
	case QtFatalMsg:
		fprintf(stderr, "Fatal: %s\n", msg);
		abort();
	}
}

int main(int argc, char *argv[]) {
	/* Definitions for slope */
	ifstream slopefile ("/usr/local/chronos/slope.conf");
	string line;
	double exec_slope;

	/* Definitions for console mode */
	QString *file, *sched;
	char *file_name = NULL, *sched_name = NULL;
	char c, optstring[]= "k:c:e:f:i:l:r:s:bdghmnopvxz";
	bool lock_flag = false, verbose_flag = false, log_flag = false,
	     no_run = false, excel_flag = false,
	     hua_flag = false, deadlock_flag = false, pi_flag = false, 
	     file_set = false, sched_set = false, cmd = false, batch_flag = false,
	     gnu_flag = false, nested_flag = false;
	int run_time = 0, cpu_usage = 0, lock_len = 0, end_usage = 0, interval = 0, main_cpu = 0;
        int exec_time=0;
	// Minimum 8 have to be there.
	if (argc < 8 && argc != 1)
	{
		cout << endl
		     << "Error:: Provide the scheduler, cpu usage,  run time and filename for taskset "
		     << endl;
		print_usage();
		return 0;
	}
	else if(argc > 8)
		cmd = true;

	/* Handle command line arguments */
	while ((c = getopt (argc, argv, optstring)) != -1)
	{
		switch (c) {
		case 'b' :
			batch_flag = true;
			break;

		case 's' :
			sched_name = optarg;
			sched_set = true;
			break;

		case 'f' :
			file_name = optarg;
			file_set = true;
			break;

		case 'c' :
			cpu_usage = atoi(optarg);
			break;

		case 'e' :
			end_usage = atoi(optarg);
			break;
			
		case 'i' :
			interval = atoi(optarg);
			break;

		case 'l' :
			lock_flag = true;
			lock_len = atoi(optarg);			
			break;

		case 'r' :
			run_time = atoi(optarg);
			break;

		case 'p' :
			pi_flag = true;
			break;

		case 'g' :
			gnu_flag = true;
			break;

		case 'h' :
			hua_flag = true;
			break;

		case 'd' :
			deadlock_flag = true;
			break;

		case 'm' :
			main_cpu = atoi(optarg);
			break;
                
                case 'k' :
                        exec_time = atoi(optarg);
			break;

		case 'n' :
			nested_flag = true;
			break;
				
		case 'v' :
			verbose_flag = true;
			break;

		case 'o' :
			log_flag = true;
			break;

		case 'x' :
			excel_flag = true;
			break;
			
		case 'z' :
			no_run = true;
			break;

		case '?' :
			if (optopt == 'f' || optopt == 'l' ||
				optopt == 'c' || optopt == 'r')
				cout << "Option -" << (char) optopt 
				     << " requires an argument" << endl;
		default :
			print_usage();
			return 0;
		}
	}

	/* If any character was used without the '-' handle them here */
	for (int index = optind; index < argc; index++)
	{
		print_usage();
		return 0;
	}

	/* If we are lacking a taskset or scheduler, die */
	if (cmd && (file_set == false || sched_set == false))
	{
		print_usage();
		return 0;
	}

	/* FIXME: If batch mode is enabled, has the end-usage and cpu-usage given? */

	file = new QString  ((const char*) file_name);
	sched = new QString ((const char*) sched_name);

	if (slopefile.is_open()) {
		getline (slopefile,line);
		exec_slope = atof(line.c_str());
		slopefile.close();
	} else {
		cout << "Error: No slope found. Please run slope_test_app" << endl;
		return 1;
	}

	//Configure Error Handler
	qInstallMsgHandler(qtMessageHandler);

	// Lock the memory space
	if(mlockall(MCL_CURRENT|MCL_FUTURE) != 0) {
		cout << "Error: Unable to lock the memory space" << endl;
		cout << "Are you running as root?" << endl;
		return 1;
	}

	//Configure the GUI
	if(!cmd) {
		QApplication app(argc, argv);

		Ui_MainWindow *ui = new Ui_MainWindow();
		QMainWindow *mw = new QMainWindow();
		QFileDialog *fd = new QFileDialog(mw, "Select Taskset", QDir::currentPath());
		ui->setupUi(mw);
		set_version(ui);
		QObject::connect(ui->changeTaskSetButton, SIGNAL(clicked()), fd, SLOT(exec()));

		//Configure the tester
		RtTester *tester = new RtTester(ui, fd);
		tester->set_slope(exec_slope);
		QObject::connect(fd, SIGNAL(fileSelected(const QString&)), tester, SLOT(fileSelected(const QString&)));
		QObject::connect(ui->batchBox, SIGNAL(toggled(bool)), tester, SLOT(changeText(bool)));
		QObject::connect(ui->lockingCBox, SIGNAL(toggled(bool)), tester, SLOT(onLockingDisable(bool)));
		mw->show();

		return app.exec();
	} else {
		RtTester *tester = new RtTester();
		tester->set_slope(exec_slope);
		int lcm = tester->fileSelected(*file);

		if (no_run == true)
			cout << "Hyper-period is: " << lcm << "s" << endl;

		if (no_run == false) {
			tester->setSchedAlgo(*sched, deadlock_flag, pi_flag, hua_flag);
			tester->setParamters(cpu_usage, end_usage, interval, exec_time, run_time, batch_flag);
			tester->setLocking(lock_flag, nested_flag, lock_len);
			tester->setOutputFormat(verbose_flag, log_flag, excel_flag, gnu_flag);
			tester->setMain(main_cpu);
			tester->startRun();
		} else {
				cout << "no run (-z) flag enabled" << endl;
		}
		
		return 0;
	}

}

