/***************************************************************************
 *   Copyright (C) 2009-2010 Virginia Tech Real-Time Systems Lab
 *   Written by Matthew Dellinger
 *   mdelling@vt.edu
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef RT_TEST_H
#define RT_TEST_H

#include <QObject>
#include <QDebug>
#include <QFileDialog>
#include <QString>
#include <QMessageBox>
#include <QDateTime>

#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <math.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <chronos/atomic_int.h>
#include "rt_sched_list.h"
#include "rt_task.h"
#include "ui_Sched_Test_App.h"

/* Priorities */
#define MAIN_PRIO			48
#define TASK_CREATE_PRIO		46
#define TASK_START_PRIO			44
#define TASK_CLEANUP_PRIO		42
#define TASK_RUN_PRIO			40

class Task;

using namespace std;

/* Tester Class */
class RtTester : public QObject
{
	Q_OBJECT

	public:

	RtTester() {
		selector_ = new SchedSelector(TASK_RUN_PRIO);
		mw_ = NULL;
		fd_ = NULL;

		InitTester();
	}

	RtTester(Ui_MainWindow *window, QFileDialog *dialog) {
		selector_ = new SchedSelector(TASK_RUN_PRIO);
		mw_ = window;
		fd_ = dialog;
		QObject::connect(mw_->runButton, SIGNAL(clicked()), this,
				 SLOT(runClicked()));
		QObject::connect(mw_->clearButton, SIGNAL(clicked()),
				 mw_->textEdit, SLOT(clear()));

		/* Add Scheduling Algos to the combo box */
		for(int i = 0; i < selector_->count(); i++) {
			mw_->schedCmbBox->addItem(selector_->AlgoName(i));
		}

		InitTester();
	}

	~RtTester() {
		delete domain_mask_;
		delete selector_;
	}

	void InitTester() {
		oss = new ostringstream(ostringstream::out);

#ifdef _SC_NPROCESSORS_ONLN
		proc_count_ = sysconf(_SC_NPROCESSORS_ONLN);
		if(proc_count_ < 2 && mw_)
			mw_->mainCPUCmbBox->setEnabled(false);
		else if(mw_)
		{
			for(int i = 0; i < proc_count_; i++)
			{
				mw_->mainCPUCmbBox->addItem(QString::number(i));
			}
		}

		domain_mask_ = new unsigned long[proc_count_];
#endif
		// Add the abort device
		if(init_aborts(&aborts_)) {
			cout << "Error: Unable to mmap abort device" << endl;
			cout << "Are you running as root?" << endl;
		}
	}
	void ClearCounters() {
		/* Performance statistics functions */
		sys_total_release_ = 0;
		sys_met_release_ = 0;
		sys_total_util_ = 0;
		sys_met_util_ = 0;
		sys_abort_count_ = 0;
		max_tardiness_ = 0;

		/* Stuff related to starting tasks correctly */
		atomic_int_init(&started_);
		barrier_ = 1;
	}

	/* Getters for local members*/
	chronos_mutex_t *lock(int lock) const { return lock_list[lock]; }
	int sys_total_release() const { return sys_total_release_; }
	int sys_total_util() const { return sys_total_util_; }
	int sys_met_release() const { return sys_met_release_; }
	int sys_met_util() const { return sys_met_util_; }
	int sys_abort_count() const { return sys_abort_count_; }
	long max_tardiness() const { return -max_tardiness_; }
	double dsr_per() const { return (double)sys_met_release()/(double)sys_total_release(); }
	double aur_per() const { return (double)sys_met_util()/(double)sys_total_util(); }
	bool verbose() const { return verbose_; }
	bool locking() const { return locking_; }
	bool nested_locking() const { return nested_locking_; }
	bool use_hua() const { return use_hua_; }
	int main_cpu() const { return main_cpu_; }
	unsigned int run_time_s() const { return run_time_; }
	unsigned int run_time_us() const { return run_time_*MILLION; }
	unsigned long lock_time() const { return lock_time_; }
	int lock_count() const { return lock_count_; }
	double usage_abs() const { return ((double)cpu_usage_)/100; }
	int usage_per() const { return cpu_usage_; }
        int usage_exec_time() const { return exec_time_; }
	double cs_len() const { return ((double)cs_len_)/100; }
	double slope() const { return slope_; }
	struct timespec * getStartTime() { return &g_start_time_; }
	struct timespec * getEndTime() { return &g_end_time_; }
	int chronos() const { return selector_->Sched() != -1; }
	int barrier() const { return barrier_; }
	int started() { return atomic_int_get(&started_); }

	/* Setters */
	void set_lock_time(unsigned long u) { lock_time_ = u; }
	void set_slope(double slope) { slope_ = slope; }
	void mark_started() { atomic_int_inc(&started_); }

	/* Non-local Getters */
	unsigned long MainCpuMask() const { return pow(2, main_cpu_); }
	int SchedAlgo() const { return selector_->Sched(); }
	QString CurrentAlgoName() const { return selector_->CurrentAlgoName(); }
	int SchedPriority() const{ return selector_->SchedPriority(); }
	int CountTasks() const { return task_list.count(); }
      
	void AddCpu(int cpu, unsigned long *mask) { *mask |= (unsigned long)pow(2, cpu); }
	Task *FindByTid(int tid);
	/* Console-only functions to set everything up */
	void fileSelectedEx(QString *file, bool no_run);
	void setSchedAlgo(QString sched, bool dead_flag, bool pi_flag, bool hua_flag);
	void setParamters(int cpu_x, int end_x, int interval, int exec_time, int run_x, bool batch_flag);
	void setOutputFormat(bool verbose_flag, bool log_flag, bool excel_flag,
			     bool gnu_flag);
	void setLocking(bool lock_flag, bool nested, int cs_len);
	void setMain(int main);
	void startRun();
         int CalcHighFreq();
         int *CalcSlowdownFactors();
	/* Main stream for logging */
	ostringstream *oss;

	public slots:
	void runClicked();
	void changeText(bool value);
	int fileSelected(const QString &file);
	void onLockingDisable(bool locking);

	private:
	/* Function Declarations */
        
	int run();
	void initReference();
	void calcLockTime();
	void initialize();
	void extract();
	bool verify();
	void print()
	{
		if(foutput.is_open())
			foutput << oss->str();
		if(mw_) {
	        	mw_->textEdit->moveCursor(QTextCursor::End, QTextCursor::MoveAnchor);
			mw_->textEdit->insertPlainText(QString(oss->str().c_str()));
		}
		else
			cout << oss->str();
		oss->str("");
	}

	/* Members */
	Ui_MainWindow *mw_;
	QFileDialog *fd_;
	SchedSelector *selector_;
	int barrier_;
	atomic_int_t started_;

	/* CPU Affinity */
	int proc_count_;
	int main_cpu_;
	unsigned long *domain_mask_;

	/* CPU Usage */
	int cpu_usage_;
	int cpu_usage_end_;
	int cpu_usage_it_;
        int exec_time_;

	/* Other run options */
	int run_time_;
	int cs_len_;
	int lock_count_;
	unsigned long lock_time_;
	bool locking_;
	bool nested_locking_;
	bool verbose_;
	bool excel_;
	bool gnuplot_;
	bool batch_mode_;
	bool use_hua_;
	bool log_to_file_;

	/* System slope */
	double slope_;

	/* Abort device */
	chronos_aborts_t aborts_;

	/* Run statistics */
	struct timespec g_start_time_;
	struct timespec g_end_time_;

	/* Task and lock lists */
	QList<Task *> task_list;
	QList<chronos_mutex_t *> lock_list;

	/* System level performance statistics */
	int sys_total_release_;// The total number of tasks
	int sys_met_release_;	// The number of tasks that met deadlines
	int sys_total_util_;	// The total utility of all tasks
	int sys_met_util_;	// The total utility of all tasks that met deadlines
	int sys_abort_count_;	// The number of threads aborted
	long max_tardiness_;	// The highest tardiness of any task

	ofstream foutput;
    };

#endif
