/* include/linux/chronos_sched.h
 *
 * Functions for the management of global lists, scheduler lists, architecture
 * functions, and mapping functions.
 *
 * Copyright (C) 2009-2010 Virginia Tech Real Time Systems Lab
 * Written by Matthew Dellinger
 */

#ifndef _CHRONOS_SCHED_H
#define _CHRONOS_SCHED_H

#ifdef CONFIG_CHRONOS

#include <linux/list.h>
#include <linux/chronos_util.h>
#include <linux/chronos_types.h>

#ifdef CONFIG_CHRONOS_SCHED_STATS
# define cschedstat_inc(rq, field)	do { (rq)->field++; } while (0)
# define cschedstat_add(rq, field, amt)	do { (rq)->field += (amt); } while (0)
# define cschedstat_set(var, val)	do { var = (val); } while (0)
#else /* !CONFIG_CHRONOS_SCHED_STATS */
# define cschedstat_inc(rq, field)	do { } while (0)
# define cschedstat_add(rq, field, amt)	do { } while (0)
# define cschedstat_set(var, val)	do { } while (0)
#endif

#define is_realtime(t) 		((t)->policy == SCHED_CHRONOS)

DECLARE_PER_CPU(struct task_struct *, global_task);
extern struct rt_sched_local fifo;

extern struct list_head rt_sched_list;
extern rwlock_t rt_sched_list_lock;
extern struct list_head global_domain_list;
extern rwlock_t global_domain_list_lock;

inline void chronos_init_cpu(int cpu);
inline int count_global_cpus(struct global_sched_domain *g);

inline int get_global_chronos_prio(struct global_sched_domain *g);
inline int get_global_chronos_sys_prio(struct global_sched_domain *g);

static inline struct task_struct * task_of_rtinfo(const struct rt_info *task)
{
	return container_of(task, struct task_struct, rtinfo);
}

static inline struct rt_info *local_task(struct list_head *task)
{
	return list_entry(task, struct rt_info, task_list[0]);
}

static inline struct rt_info *global_task(struct list_head *task)
{
	return list_entry(task, struct rt_info, task_list[1]);
}

/* Returns 1 if a scheduler is global */
static inline int is_global(struct sched_base *sched)
{
	return (sched->id & SCHED_GLOBAL_MASK) ? 1 : 0;
}

/* Returns 1 if a task is in the global list */
static inline int in_global_list(struct rt_info *task)
{
	return (!list_empty(&task->task_list[1]));
}

static inline void mark_for_global_insert(struct rt_info *r, struct global_sched_domain *g)
{
	if(g) {
		atomic_inc(&g->tasks);
		r->flags |= TASK_FLAG_INSERT_GLOBAL;
	}
}

static inline void clear_global_insert(struct rt_info *r)
{
	r->flags &= ~TASK_FLAG_INSERT_GLOBAL;
}

inline void lock_global_task_list(struct global_sched_domain *g);
inline void unlock_global_task_list(struct global_sched_domain *g);

inline int trylock_global_sched_lock(struct global_sched_domain *g);
inline void lock_global_sched_lock(struct global_sched_domain *g);
inline void unlock_global_sched_lock(struct global_sched_domain *g);

/* Add a local real-time scheduler */
int add_local_scheduler(struct rt_sched_local *scheduler);
void remove_local_scheduler(struct rt_sched_local *scheduler);

/* Add a global real-time scheduler */
int add_global_scheduler(struct rt_sched_global *scheduler);
void remove_global_scheduler(struct rt_sched_global *scheduler);

void add_scheduler_nocheck(struct sched_base *scheduler, int is_global);

/* Find a scheduler */
struct rt_sched_local * get_local_scheduler(int scheduler);
struct rt_sched_global * get_global_scheduler(int scheduler);

/* Add and remove a task from the global scheduling queue */
inline int has_global_tasks(struct global_sched_domain *g);
inline int global_tasks(struct global_sched_domain *g);
inline void _remove_task_global(struct rt_info *task, struct global_sched_domain *g);
void test_add_task_global(struct rt_info *task, struct global_sched_domain *g);
void test_remove_task_global(struct rt_info *task, struct global_sched_domain *g);
void check_global_insert(struct task_struct *t, struct global_sched_domain *g);

/* Add and remove a domain from the list */
struct global_sched_domain *create_global_domain(struct rt_sched_global *g, int prio);
void add_global_domain(struct global_sched_domain *domain);
void remove_global_domain(struct global_sched_domain *domain);
void print_global_domain(struct global_sched_domain *domain, struct seq_file *m);
void print_global_domains(struct seq_file *m);

void build_list(struct rt_info *head, int list, int cpus);
void build_list_array (struct rt_info **head, int cpus);

/* Mapping functions */
void generic_map_all_tasks(struct rt_info *best, struct global_sched_domain *g);
void map_to_me(struct rt_info *best, struct global_sched_domain *g);

/* Architecture init functions */
int init_concurrent(struct global_sched_domain *g, int block);
int init_stw(struct global_sched_domain *g, int block);

/* Architecture release functions */
void release_concurrent(struct global_sched_domain *g);
void release_generic(struct global_sched_domain *g);
#define release_stw release_generic

struct rt_info * presched_stw_generic(struct list_head *head);
struct rt_info * presched_concurrent_generic(struct list_head *head);
struct rt_info * presched_abort_generic(struct list_head *head);

extern struct rt_sched_arch rt_sched_arch_concurrent;
extern struct rt_sched_arch rt_sched_arch_stw;

#endif//CONFIG_CHRONOS
#endif

