/* include/linux/chronos_types.h
 *
 * All #defines and data structures used in ChronOS
 *
 * Author(s)
 *	- Matthew Dellinger, mdelling@vt.edu
 *
 * Copyright (C) 2009-2011 Virginia Tech Real Time Systems Lab
 */

#ifndef _CHRONOS_TYPES_H
#define _CHRONOS_TYPES_H

#include <linux/time.h>
#include <asm/atomic.h>
#include <asm/mcs.h>

/* Task Flags - all others to be used in algorithm-specific ways */
#define TASK_FLAG_NONE			0x00
#define TASK_FLAG_MASK			0xFF
#define TASK_FLAG_ABORTED		0x01
#define TASK_FLAG_HUA			0x02
#define TASK_FLAG_SCHEDULED		0x04
#define TASK_FLAG_DEADLOCKED		0x08
#define TASK_FLAG_INSERT_GLOBAL		0x80

/* Masks for getting the scheduler from userspace
 * 1 bit - global
 * 23 bits - scheduler number
 * 8 bits - flags
 */
#define SCHED_GLOBAL_MASK		0x80
#define SCHED_NUM_MASK			0x7F
#define SCHED_FLAGS_MASK		0xFF

/* Scheduler - behaviors for each flag may or may not be defined */
#define SCHED_RT_FIFO 			0x00
#define SCHED_RT_RMA 			0x01
#define SCHED_RT_EDF 			0x02
#define SCHED_RT_HVDF			0x03
#define SCHED_RT_LBESA			0x04
#define SCHED_RT_DASA_ND		0x05
#define SCHED_RT_DASA			0x06
#define SCHED_RT_FIFO_RA		0x07
#define SCHED_RT_BASE_EDF               0x08
#define SCHED_RT_STATIC_EDF             0x09
#define SCHED_RT_CC_EDF                 0x0A
#define SCHED_RT_LAEDF                  0x0B
#define SCHED_RT_SNOWDON_MIN            0x0C
#define SCHED_RT_DRA                    0x0D
#define SCHED_RT_DRA_OTE                0x0E
#define SCHED_RT_AGR1                   0x0F
#define SCHED_RT_AGR2                   0x10
#define SCHED_RT_HS                     0x11
#define SCHED_RT_DS                     0x12
#define SCHED_RT_USFI_EDF               0x13
#define SCHED_RT_REUA                   0x14
#define SCHED_RT_GFIFO			0x80
#define SCHED_RT_GRMA			0x81
#define SCHED_RT_GEDF			0x82
#define SCHED_RT_GNP_EDF		0x83
#define SCHED_RT_GHVDF			0x84
#define SCHED_RT_GMUA			0x85
#define SCHED_RT_GGUA			0x86
#define SCHED_RT_NGGUA			0x87
#define SCHED_RT_GNP_HVDF		0x88

/* Scheduling Flags */
/* PI == Priority Inheritance
 * HUA == HUA style abort handlers
 * ABORT_IDLE == abort the task only when the system is idle
 * NO_DEADLOCKS == with deadlock prevention
 *
 * Also, 0x*0 is left open for future flags.
 */
#define SCHED_FLAG_NONE			0x00
#define SCHED_FLAG_MASK			0xFF
#define SCHED_FLAG_HUA			0x01
#define SCHED_FLAG_PI			0x02
#define SCHED_FLAG_NO_DEADLOCKS		0x04

/* Number of lists for schedules */
#define SCHED_LISTS			4

/* Sorting Keys */
#define SORT_KEY_NONE			0
#define SORT_KEY_DEADLINE		1
#define SORT_KEY_PERIOD			2
#define SORT_KEY_LVD			3
#define SORT_KEY_GVD			4
#define SORT_KEY_TDEADLINE		5

/* Struct for a owner-tracking futex
 * USERSPACE SHARED
 */
struct mutex_data {
	u32 value;
	int owner;
	unsigned long id;
};

struct mutex_head {
	struct list_head list;
	struct rt_info *owner_t;
	struct mutex_data *mutex;
	unsigned long id;
};

struct list_entry {
	struct rt_info *next;
	struct rt_info *prev;
};

struct abort_info {
	struct timespec deadline;
	unsigned long exec_time;
	int max_util;
};

/* Struct for passing parameters down to kernel
 * USERSPACE SHARED
 */
struct rt_data {
	int tid;
	int prio;
	unsigned long exec_time;
        int util_taskset;
        unsigned long locked_usage;
        int mutex_flag;
        int high_freq;
	int max_util;
	struct timespec *deadline;
	struct timespec *period;
        struct timespec *buffer1;
        unsigned long buffer2;
        unsigned long buffer3;
        struct mutex_data *mutex0;
        struct mutex_data *mutex1;
        int *n_freq;
        int n_frequency[20];
        
         
};

/* Structure used by x-GUA class of algos for the DAG */
struct rt_graph {
	struct timespec agg_left;
	unsigned long 	agg_util;
	long in_degree;
	long out_degree;
	struct rt_info *neighbor_list;
	struct rt_info *next_neighbor;
	struct rt_info *parent;
	struct rt_info *depchain;
};

struct alpha_member {
        int i;
        int i_end;
	unsigned long deadline;
        long left;
        long rem_time;
        unsigned long wcet_can;
      struct list_head list;
}; 



/* Structure attached to struct task_struct */
struct rt_info {
	/* Real-Time information */
	struct timespec deadline;		/* monotonic time */
	struct timespec temp_deadline;		/* monotonic time */
	struct timespec period;			/* relative time */
	struct timespec left;			/* relative time */
	unsigned long exec_time;		/* WCET, us */
	int max_util;
	long local_ivd;
	long global_ivd;
	unsigned int seg_start_us;
        struct alpha_member a_member;
       
        int util_taskset;
        long load;
        long prev_load;
        int total_util;
        int count_for_cycles_div;
        int flag_begin_end_check;
        int flag_check_rtseg;
        unsigned long old_task_time;
        unsigned long budget;
        long buffer2,buffer3,buffer4;
        struct mutex_data *mutex0;
        struct mutex_data *mutex1;
        int mutex_flag;
        int high_freq;
        int *n_freq;
        int n_frequency[20];
        unsigned long locked_usage;
             

	/* Lists FIXME: convert to array of list_heads*/  
	/* 0 is local, 1 is global... should fix to pound defines */
	struct list_head task_list[2];
	struct list_entry list[SCHED_LISTS];

	/* DAG used by x-GUA class of algorthims */
	struct rt_graph graph;

	/* Lock information */
	struct mutex_head *requested_resource;
	struct rt_info *dep;

	/* Abort information */
	struct abort_info abortinfo;

	/* Task state information */
	unsigned char flags;
	int cpu;
};

struct global_sched_domain {
	/* The global scheduler */
	struct rt_sched_global *scheduler;
	/* The global task list */
	struct list_head global_task_list;
	/* The CPUs in this domain */
	cpumask_t global_sched_mask;
	/* Global scheduling priority in this domain */
	int prio;
	/* Task list lock */
	raw_spinlock_t global_task_list_lock;
	/* Scheduling lock */
	mcs_lock_t global_sched_lock;
	/* The head node for the mapping list... FIXME: ugly solution */
	struct rt_info *mapping_list;
	/* Global domain list */
	struct list_head list;
	/* Current task count */
	atomic_t tasks;
};

struct rt_sched_arch {
	int (*arch_init) (struct global_sched_domain *g, int block);
	void (*arch_release) (struct global_sched_domain *g);
	void (*map_tasks) (struct rt_info *head, struct global_sched_domain *g);
};

struct sched_base {
	struct list_head list;
	/* Scheduler Name */
	char *name;
	/* Scheduling number and flags */
	unsigned int id;
	/* Sort key for the global list */
	unsigned int sort_key;
	/* The mask of CPUs this scheduler is active on */
	cpumask_t active_mask;
};

struct rt_sched_local {
	/* Base information */
	struct sched_base base;
	/* Flags - currently not needed for any globals */
	unsigned int flags; // should maybe be per-processor?
	/* Scheduling function */
	struct rt_info* (*schedule) (struct list_head *head, int flags);
};

struct rt_sched_global {
	/* Base information */
	struct sched_base base;
	/* Scheduling functions */
	struct rt_info* (*schedule) (struct list_head *head, struct global_sched_domain *g);
	struct rt_info* (*preschedule) (struct list_head *head);
	struct rt_sched_arch *arch;
	/* The local scheduler to be used with this global */
	int local;
};
#endif

