/* kernel/chronos_sched.c
 *
 * Functions for the management of global lists, scheduler lists, architecture
 * functions, and mapping functions.
 *
 * Author(s):
 *	- Matthew Dellinger, mdelling@vt.edu
 *
 * Copyright (C) 2009-2011 Virginia Tech Real Time Systems Lab
 */

#include <linux/chronos_sched.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <asm/atomic.h>
#include <asm/mcs.h>

/* List of all the real-time scheduling algorithms in the system */
LIST_HEAD(rt_sched_list);
DEFINE_RWLOCK(rt_sched_list_lock);
LIST_HEAD(global_domain_list);
DEFINE_RWLOCK(global_domain_list_lock);

/* The task_struct pointers the global scheduler fills */
DEFINE_PER_CPU(struct task_struct *, global_task);
/* The MCS lock node for the global scheduling lock */
DEFINE_PER_CPU(mcs_node_t, global_sched_lock_node);

inline void chronos_init_cpu(int cpu)
{
	mcs_node_init(&per_cpu(global_sched_lock_node, cpu));
	per_cpu(global_task, cpu) = NULL;
}

/* FIFO, just so that by default we don't muck with Linux */
struct rt_info * sched_fifo (struct list_head *head, int flags) {
	return local_task(head->next);
}

struct rt_sched_local fifo = {
	.base.name = "FIFO",
	.base.id = 0,
	.base.sort_key = SORT_KEY_NONE,
	.flags = 0,
	.schedule = sched_fifo
};

/* All the prio functions can be called without knowing if we have a valid domain
 * such as in sched_setscheduler. Hence we check g.
 */
inline int get_global_chronos_prio(struct global_sched_domain *g)
{
	return g ? g->prio : 0;
}

inline int get_global_chronos_sys_prio(struct global_sched_domain *g)
{
	return g ? MAX_RT_PRIO - g->prio - 1 : MAX_RT_PRIO;
}

/* The cpus the global scheduler is using */
inline int count_global_cpus(struct global_sched_domain *g)
{
	return cpumask_weight(&g->global_sched_mask);
}
EXPORT_SYMBOL(count_global_cpus);

/* Lock and unlock the task list */
inline void lock_global_task_list(struct global_sched_domain *g)
{
	raw_spin_lock(&g->global_task_list_lock);
}

inline void unlock_global_task_list(struct global_sched_domain *g)
{
	raw_spin_unlock(&g->global_task_list_lock);
}

/* Lock or unlock the global scheduling lock */
inline int is_locked_global_sched_lock(struct global_sched_domain *g)
{
	return mcs_is_locked(&g->global_sched_lock);
}

inline int trylock_global_sched_lock(struct global_sched_domain *g)
{
	return mcs_trylock(&g->global_sched_lock, &per_cpu(global_sched_lock_node, raw_smp_processor_id()));
}

inline void lock_global_sched_lock(struct global_sched_domain *g)
{
	mcs_lock(&g->global_sched_lock, &per_cpu(global_sched_lock_node, raw_smp_processor_id()));
}

inline void unlock_global_sched_lock(struct global_sched_domain *g)
{
	mcs_unlock(&g->global_sched_lock, &per_cpu(global_sched_lock_node, raw_smp_processor_id()));
}

/* Scheduler list management */
/* Actually add the scheduler. Add at different ends for readability */
void add_scheduler_nocheck(struct sched_base *scheduler, int is_global)
{
	write_lock(&rt_sched_list_lock);
	if(is_global)
		list_add_tail(&scheduler->list, &rt_sched_list);
	else
		list_add(&scheduler->list, &rt_sched_list);

	write_unlock(&rt_sched_list_lock);
}

static int add_rt_scheduler(struct sched_base *scheduler, int is_global)
{
	if(!list_empty(&scheduler->list))
		return -EEXIST;

	cpumask_clear(&scheduler->active_mask);
	add_scheduler_nocheck(scheduler, is_global);

	return 0;
}

int add_local_scheduler(struct rt_sched_local *scheduler)
{
	return add_rt_scheduler(&scheduler->base, 0);
}
EXPORT_SYMBOL(add_local_scheduler);

int add_global_scheduler(struct rt_sched_global *scheduler)
{
	return add_rt_scheduler(&scheduler->base, 1);
}
EXPORT_SYMBOL(add_global_scheduler);

static void remove_scheduler(struct sched_base *scheduler)
{
	/* Set scheduler to FIFO to prevent problems */
 	set_scheduler_mask(&fifo, NULL, &scheduler->active_mask, 0);

	write_lock(&rt_sched_list_lock);
	list_del_init(&scheduler->list);
	write_unlock(&rt_sched_list_lock);
}

void remove_local_scheduler(struct rt_sched_local *global) {
	remove_scheduler(&global->base);
}
EXPORT_SYMBOL(remove_local_scheduler);

void remove_global_scheduler(struct rt_sched_global *global) {
	remove_scheduler(&global->base);
}
EXPORT_SYMBOL(remove_global_scheduler);

static struct sched_base * get_scheduler(int scheduler)
{
	struct sched_base *it, *found = NULL;

	read_lock(&rt_sched_list_lock);
	list_for_each_entry(it, &rt_sched_list, list) {
		if(it->id == scheduler) {
			found = it;
			break;
		}
	}
	read_unlock(&rt_sched_list_lock);

	return found;
}

struct rt_sched_local * get_local_scheduler(int scheduler) {
	struct sched_base *base = get_scheduler(scheduler);
	if(base)
		return container_of(base, struct rt_sched_local, base);

	return NULL;
}

struct rt_sched_global * get_global_scheduler(int scheduler) {
	struct sched_base *base = get_scheduler(scheduler);
	if(base)
		return container_of(base, struct rt_sched_global, base);

	return NULL;
}

/* Global task list management
 * test_* add and remove functions can be called without knowing the validity of
 * the domain, normal add and remove functions require a valid domain
 */

/* Two ways to check for global tasks
 *
 * has_global_tasks() return whether the list is empty
 * global_tasks() returns the task count, which is more optimistic and includes
 *	tasks that may not yet have been added to the list. It can also be
 *	called on an invalid domain.
 */
inline int has_global_tasks(struct global_sched_domain *g)
{
	return !list_empty(&g->global_task_list);
}

inline int global_tasks(struct global_sched_domain *g)
{
	return g ? atomic_read(&g->tasks) : 0;
}

inline void _add_task_global(struct rt_info *task, struct global_sched_domain *g)
{
	insert_on_global_queue(task, &g->global_task_list, g->scheduler->base.sort_key);
}

void add_task_global(struct rt_info *task, struct global_sched_domain *g)
{
	lock_global_task_list(g);
	_add_task_global(task, g);
	unlock_global_task_list(g);
}

void test_add_task_global(struct rt_info *task, struct global_sched_domain *g)
{
	if(g)
		add_task_global(task, g);
}

inline void _remove_task_global(struct rt_info *task, struct global_sched_domain *g)
{
	list_del_init(&task->task_list[1]);
	atomic_dec(&g->tasks);
}
EXPORT_SYMBOL(_remove_task_global);

void remove_task_global(struct rt_info *task, struct global_sched_domain *g)
{
	lock_global_task_list(g);
	_remove_task_global(task, g);
	unlock_global_task_list(g);
}

void test_remove_task_global(struct rt_info *task, struct global_sched_domain *g)
{
	if(g && in_global_list(task))
		remove_task_global(task, g);
}

/* Check if a task needs inserted on the global list for a given domain */
void check_global_insert(struct task_struct *p, struct global_sched_domain *g)
{
	if((p->policy == SCHED_CHRONOS) && (p->rtinfo.flags & TASK_FLAG_INSERT_GLOBAL)) {
		test_add_task_global(&p->rtinfo, g);
		clear_global_insert(&p->rtinfo);
	}
}

/* Global domain management */
struct global_sched_domain *create_global_domain(struct rt_sched_global *g, int prio)
{
	struct global_sched_domain *domain;

	domain = kmalloc(sizeof(struct global_sched_domain), GFP_KERNEL);

	if(domain) {
		INIT_LIST_HEAD(&domain->global_task_list);
		INIT_LIST_HEAD(&domain->list);
		raw_spin_lock_init(&domain->global_task_list_lock);
		mcs_lock_init(&domain->global_sched_lock);
		atomic_set(&domain->tasks, 0);
		domain->scheduler = g;
		domain->prio = prio;
	} else
		printk("Failed creating global scheduling domain with %s\n", g->base.name);


	return domain;
}

void add_global_domain(struct global_sched_domain *domain)
{
	int cpu;

	write_lock(&global_domain_list_lock);
	list_add(&domain->list, &global_domain_list);
	write_unlock(&global_domain_list_lock);
	printk("Added global scheduling domain with %s\n", domain->scheduler->base.name);
	printk("\t[ ");
	for_each_cpu_mask(cpu, domain->global_sched_mask) {
		printk("%i ", cpu);
	}
	printk("]\n");
}

void remove_global_domain(struct global_sched_domain *domain)
{
	write_lock(&global_domain_list_lock);
	list_del(&domain->list);
	write_unlock(&global_domain_list_lock);
}

/* Global list building function
 *
 * This takes a list built on the rt_info->list pointers and sets it up to be
 * passed to a mapping function.
 */
void build_list(struct rt_info *head, int list, int cpus)
{
	if(!head)
		return;

	/* The final list is always built on list 0 */
	trim_list(head, list, cpus);

	if(list)
		copy_list(head, list, 0);
}
EXPORT_SYMBOL(build_list);

/* Global list building function
 *
 * This takes a array of tasks and sets it up to be passed to a mapping function.
 *
 * Returns 1 if the list length is greater than the number of processors
 * Otherwise, returns 0
 */
void build_list_array(struct rt_info **head, int cpus)
{
	int i;
	struct rt_info *curr = NULL, *first = head[0];

	if(!head || !first)
		return;

	initialize_lists(first);

	for(i = 1; i < cpus; i++) {
		curr = head[i];
		if (curr) {
			curr->list[0].next = first->list[0].next;
			curr->list[0].prev = first;
			curr->list[0].next->list[0].prev = curr;
			first->list[0].next = curr;
		} else
			break;
	}
}
EXPORT_SYMBOL(build_list_array);

/* Three different ways to IPI other cpus */
static void reschedule_count_global_cpus(struct global_sched_domain *g, int prio, int tasks)
{
	int count = tasks, cpu;
	cpumask_t mask;

	cpumask_copy(&mask, &g->global_sched_mask);
	cpumask_clear_cpu(raw_smp_processor_id(), &mask);

	for_each_cpu_mask(cpu, mask) {
		if(prio_resched_cpu(cpu, prio))
			count--;
		if(count == 0)
			return;
	}
}

static void reschedule_trycount_global_cpus(struct global_sched_domain *g, int prio, int tasks)
{
	int count = tasks, cpu;
	cpumask_t mask;

	cpumask_copy(&mask, &g->global_sched_mask);
	cpumask_clear_cpu(raw_smp_processor_id(), &mask);

	for_each_cpu_mask(cpu, mask) {
		count--;
		prio_resched_cpu(cpu, prio);
		if(count == 0)
			return;
	}
}

static void reschedule_all_global_cpus(struct global_sched_domain *g, int prio)
{
	int cpu;
	cpumask_t mask;

	cpumask_copy(&mask, &g->global_sched_mask);
	cpumask_clear_cpu(raw_smp_processor_id(), &mask);

	for_each_cpu_mask(cpu, mask) {
		prio_resched_cpu(cpu, prio);
	}
}

static struct task_struct * find_best_task(int cpu, struct global_sched_domain *g, cpumask_t *m)
{
	struct task_struct *task, *best;
	struct rt_info *curr, *head;

	if(!g->mapping_list)
		return NULL;

	curr = g->mapping_list;
	head = g->mapping_list;
	best = NULL;

	do {
		task = task_of_rtinfo(curr);

		if(task_cpu(task) == cpu) {
			best = task;
			if(task->oncpu)
				goto out;
		}

		curr = curr->list[0].next;
	} while (curr != head);

	if(!best)
		return NULL;

	curr = &(best->rtinfo);
out:
	cpumask_clear_cpu(cpu, m);
	g->mapping_list = (curr->list[0].next == curr) ? NULL : curr->list[0].next;
	list_remove(curr, 0);
	return best;
}

static struct task_struct * find_any_task(struct global_sched_domain *g)
{
	struct task_struct *task;
	struct rt_info *curr;

	curr = g->mapping_list;
	task = task_of_rtinfo(curr);

	g->mapping_list = (curr->list[0].next == curr) ? NULL : curr->list[0].next;
	list_remove(curr, 0);
	return task;
}

/* The default mapping function
 *
 * The goal is to assign n tasks to n cpus with the least migration possible.
 * This is used for algorithms like GEDF that just select the n best tasks.
 */
void generic_map_all_tasks(struct rt_info *best, struct global_sched_domain *g)
{
	int cpu;
	cpumask_t mask;

	g->mapping_list = best;
	cpumask_copy(&mask, &g->global_sched_mask);

	for_each_cpu_mask(cpu, g->global_sched_mask) {
		per_cpu(global_task, cpu) = find_best_task(cpu, g, &mask);
	}

	for_each_cpu_mask(cpu, mask) {
		if(!g->mapping_list)
			return;
		per_cpu(global_task, cpu) = find_any_task(g);
	}
}

void map_to_me(struct rt_info *best, struct global_sched_domain *g)
{
	int cpu = raw_smp_processor_id();

	if(best)
		per_cpu(global_task, cpu) = task_of_rtinfo(best);
	else
		per_cpu(global_task, cpu) = NULL;
}

/* Block on the global scheduling lock */
static void block_generic(struct global_sched_domain *g)
{
	if(is_locked_global_sched_lock(g)) {
		lock_global_sched_lock(g);
		unlock_global_sched_lock(g);
	}
}

/* Architecture Init functions */
int init_concurrent(struct global_sched_domain *g, int block)
{
	lock_global_task_list(g);

	return 1;
}

int init_stw(struct global_sched_domain *g, int block)
{
	int prio, tasks;

	if(block == 2 || !trylock_global_sched_lock(g)) {
		block_generic(g);
		return 0;
	}

	tasks = atomic_read(&g->tasks);
	prio = get_global_chronos_sys_prio(g);

	if(tasks <= count_global_cpus(g))
		reschedule_trycount_global_cpus(g, prio + 1, tasks);
	else
		reschedule_all_global_cpus(g, prio);

	lock_global_task_list(g);

	return 1;
}

/* Architecture release functions */
void release_concurrent(struct global_sched_domain *g)
{
	int prio, tasks;

	unlock_global_task_list(g);

	tasks = atomic_read(&g->tasks);
	prio = get_global_chronos_sys_prio(g) + 1;

	if(tasks < count_global_cpus(g) - 1)
		reschedule_count_global_cpus(g, prio, tasks);
	else
		reschedule_all_global_cpus(g, prio);
}

void release_generic(struct global_sched_domain *g)
{
	unlock_global_sched_lock(g);
	unlock_global_task_list(g);
}

struct rt_info * presched_stw_generic(struct list_head *head)
{
	return NULL;
}
EXPORT_SYMBOL(presched_stw_generic);

struct rt_info * presched_abort_generic(struct list_head *head)
{
	struct rt_info *it;

	list_for_each_entry(it, head, task_list[0]) {
		if(check_task_abort_nohua(it))
			return it;
	}

	return NULL;
}
EXPORT_SYMBOL(presched_abort_generic);

struct rt_info * presched_concurrent_generic(struct list_head *head)
{
	int cpu = raw_smp_processor_id();
	struct rt_info *it;

	list_for_each_entry(it, head, task_list[0]) {
		if(it->cpu == cpu)
			return it;
	}

	return NULL;
}
EXPORT_SYMBOL(presched_concurrent_generic);

struct rt_sched_arch rt_sched_arch_concurrent = {
	.arch_init = init_concurrent,
	.arch_release = release_concurrent,
	.map_tasks = map_to_me
};
EXPORT_SYMBOL(rt_sched_arch_concurrent);

struct rt_sched_arch rt_sched_arch_stw = {
	.arch_init = init_stw,
	.arch_release = release_stw,
	.map_tasks = generic_map_all_tasks
};
EXPORT_SYMBOL(rt_sched_arch_stw);

